#!/usr/bin/env python3
# Copyright (C) 2022 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Entrypoint for Python module
"""

from galene_stream.cli import main

if __name__ == "__main__":
    main()
